"""The sample demonstrates real-time data receiving"""

import logging
import threading

import win32com.client

import CELEnvironment

#region Settings

# Logging settings: specify log level and log format here.
# Set log level to DEBUG to see more internal messages.
# See more information about logging lib usage: https://docs.python.org/3/howto/logging-cookbook.html
CELEnvironment.LOG_LEVEL = logging.INFO
# CELEnvironment.LOG_FORMAT = ...

# CELEnvironment.SinkInternal uses own LOG_LEVEL (which is inherited from the global one) and own LOG_FORMAT.
# You can redefine them as follows:
# CELEnvironment.SinkInternal.LOG_LEVEL = logging.DEBUG
# CELEnvironment.SinkInternal.LOG_FORMAT = ...

LOG = CELEnvironment.GetLogger()

#endregion


def QuoteType2String(quoteType):
    return {
        win32com.client.constants.qtAsk: "Best ask",
        win32com.client.constants.qtBid: "Best bid",
        win32com.client.constants.qtCohUndAsk: "Coherent underlying price for the best ask",
        win32com.client.constants.qtCohUndBid: "Coherent underlying price for the best bid",
        win32com.client.constants.qtDayHigh: "Current day high price",
        win32com.client.constants.qtDayLow: "Current day low price",
        win32com.client.constants.qtDayOpen: "Day open price",
        win32com.client.constants.qtImpliedAsk: "Implied best ask",
        win32com.client.constants.qtImpliedBid: "Implied best bid",
        win32com.client.constants.qtIndicativeOpen: "Indicative open",
        win32com.client.constants.qtMarker: "Marker price",
        win32com.client.constants.qtOutrightAsk: "Outright best ask",
        win32com.client.constants.qtOutrightBid: "Outright best bid",
        win32com.client.constants.qtSettlement: "Settlement price",
        win32com.client.constants.qtTodayMarker: "Marker price",
        win32com.client.constants.qtTrade: "Last trade price",
        win32com.client.constants.qtYesterdaySettlement: "Yesterday's settlement price"
    }[quoteType]


class RealTimeMarketDataReceiving(CELEnvironment.CELSinkBase):
    def __init__(self):
        self.eventDone = threading.Event()

    def Init(self, celEnvironment):
        self.celEnvironment = celEnvironment

    def Start(self):
        for i in range(1, 4):
            symbol = f"CLE?{i}"
            LOG.info("Request realtime market data for %s", symbol)
            self.celEnvironment.cqgCEL.NewInstrument(symbol)

        self.eventDone.wait(100)

    def OnDataError(self, cqgError, errorDescription):
        if cqgError is not None:
            error = win32com.client.Dispatch(cqgError)
            LOG.error("OnDataError: Code: %d Description: %s", error.Code, error.Description)
        self.eventDone.set()

    def OnInstrumentResolved(self, symbol, cqgInstrument, cqgError):
        if cqgError is not None:
            error = win32com.client.Dispatch(cqgError)
            LOG.error("OnInstrumentResolved error: Error code: %d Description: %s", error.Code, error.Description)
            self.eventDone.set()
            return

        instrument = win32com.client.Dispatch(cqgInstrument)
        LOG.info("OnInstrumentResolved: Symbol: %s Instrument full name: %s", symbol, instrument.FullName)
        instrument.DataSubscriptionLevel = win32com.client.constants.dsQuotesAndBBA
        LOG.info("OnInstrumentResolved: DataSubscriptionLevel is changed")

    def OnInstrumentChanged(self, cqgInstrument, cqgQuotes, cqgInstrumentProperties):
        instrument = win32com.client.Dispatch(cqgInstrument)
        quotes = win32com.client.Dispatch(cqgQuotes)
        LOG.info("New quotes for %s", instrument.FullName)
        for quote in quotes:
            if not quote.IsValid:
                continue

            LOG.info(
                "  Type: %s Price: %f Volume: %d Time: %s",
                QuoteType2String(quote.Type),
                quote.Price,
                quote.Volume,
                quote.Timestamp
            )


if __name__ == "__main__":
    CELEnvironment.Start(RealTimeMarketDataReceiving)
