"""The sample demonstrates how to define a strategy and place a strategy order"""

import logging
import threading

import win32com.client
from win32com.client import constants

import CELEnvironment
from CELEnvironment import AssertMessage

#region Settings

# Replace it with real login before run
LOGIN = "your_gw_login"
PASSWORD = "your_gw_password"

# Replace it with enabled symbol
STRATEGY = "SPREAD(CLE-DD)"

# Events waiting time
TIMEOUT = 10

# Logging settings: specify log level and log format here.
# Set log level to DEBUG to see more internal messages.
# See more information about logging lib usage: https://docs.python.org/3/howto/logging-cookbook.html
CELEnvironment.LOG_LEVEL = logging.INFO
# CELEnvironment.LOG_FORMAT = ...

# CELEnvironment.SinkInternal uses own LOG_LEVEL (which is inherited from the global one) and own LOG_FORMAT.
# You can redefine them as follows:
# CELEnvironment.SinkInternal.LOG_LEVEL = logging.DEBUG
# CELEnvironment.SinkInternal.LOG_FORMAT = ...

LOG = CELEnvironment.GetLogger()

#endregion


class StrategyPlacingSample(CELEnvironment.CELSinkBase):
    def __init__(self):
        self.account = None
        self.eventGatewayIsUp = threading.Event()
        self.eventGatewayIsDown = threading.Event()
        self.eventInstrumentIsReady = threading.Event()
        self.eventAccountIsReady = threading.Event()
        self.eventStrategyIsReady = threading.Event()
        self.eventOrderPlaced = threading.Event()

    def Init(self, celEnvironment):
        self.celEnvironment = celEnvironment

    def createOrder(self, quantity, price):
        LOG.info("Creating a default execution pattern for limit orders...")
        executionPattern = win32com.client.Dispatch(
            self.celEnvironment.cqgCEL.CreateExecutionPattern(self.dispatchedStrategy, constants.otLimit))
        LOG.info("Creating a strategy order...")
        order = win32com.client.Dispatch(
            self.celEnvironment.cqgCEL.CreateStrategyOrder(constants.otLimit, self.dispatchedStrategy, self.account,
                                                           None,
                                                           quantity, constants.osdUndefined, price))
        LOG.info("Set an execution pattern order property...")
        properties = win32com.client.Dispatch(order.Properties)
        execPatternProperty = properties(constants.opExecutionPattern)
        execPatternProperty.Value = executionPattern.PatternString
        return order

    def Start(self):
        LOG.info("Connecting to GW")
        self.celEnvironment.cqgCEL.GWLogon(LOGIN, PASSWORD)

        LOG.info("Waiting for GW connection...")
        AssertMessage(self.eventGatewayIsUp.wait(TIMEOUT), "GW connection timeout!")

        self.celEnvironment.cqgCEL.AccountSubscriptionLevel = constants.aslAccountUpdatesAndOrders
        LOG.info("Waiting for accounts coming...")
        AssertMessage(self.eventAccountIsReady.wait(TIMEOUT), "Accounts coming timeout!")

        LOG.info("Select the first account")
        accounts = win32com.client.Dispatch(self.celEnvironment.cqgCEL.Accounts)
        self.account = win32com.client.Dispatch(accounts.ItemByIndex(0))

        LOG.info("Define strategy...")
        definedStrategy = self.celEnvironment.cqgCEL.DefineStrategyEx(STRATEGY, self.account)

        LOG.info("Waiting for strategy resolution...")
        AssertMessage(self.eventStrategyIsReady.wait(TIMEOUT), "Strategy resolution timeout!")

        self.dispatchedStrategy = win32com.client.Dispatch(definedStrategy)

        LOG.info("Subscribe to %s for a best ask value receiving. It will be used for an order placing...", STRATEGY)
        self.celEnvironment.cqgCEL.NewInstrument(STRATEGY)
        LOG.info("%s subscription waiting...", STRATEGY)
        AssertMessage(self.eventInstrumentIsReady.wait(TIMEOUT), "Instrument resolution timeout!")

        instrument = win32com.client.Dispatch(self.instrument)

        bestAsk = instrument.Ask
        AssertMessage(bestAsk.IsValid, "Error! Can't set an order price due to invalid BBA")

        LOG.info("Best ask value is %f", bestAsk.Price)
        LOG.info("Calculating a price that above the market by 10 tick sizes...")
        sellOrderPrice = bestAsk.Price + 10 * instrument.TickSize
        LOG.info("It is %f", sellOrderPrice)

        LOG.info("Starting a limit sell order creation procedure...")
        order = self.createOrder(-1, sellOrderPrice)

        LOG.info("Place the order...")
        order.Place()

        LOG.info("Waiting for order placing...")
        AssertMessage(self.eventOrderPlaced.wait(TIMEOUT), "Order placing timeout!")

        LOG.info("Logoff from GW")
        self.eventGatewayIsDown.clear()
        self.celEnvironment.cqgCEL.GWLogoff()
        AssertMessage(self.eventGatewayIsDown.wait(TIMEOUT), "GW disconnection timeout!")

        LOG.info("Done.")

    def OnGWConnectionStatusChanged(self, newStatus):
        if newStatus == constants.csConnectionUp:
            LOG.info("GW connection is UP!")
            self.eventGatewayIsUp.set()
        if newStatus == constants.csConnectionDown:
            LOG.info("GW connection is DOWN!")
            self.eventGatewayIsDown.set()

    def OnAccountChanged(self, changeType, cqgAccount, cqgPosition):
        if changeType != constants.actAccountsReloaded:
            return

        LOG.info("Accounts are ready!")
        self.eventAccountIsReady.set()

    def OnStrategyDefinitionProgress(self, cqgDefinition, cqgError):
        if cqgError is not None:
            error = win32com.client.Dispatch(cqgError)
            LOG.error("OnStrategyDefinitionProgress error: Code: %d Description: %s", error.Code, error.Description)
            return

        dispatchedDefinition = win32com.client.Dispatch(cqgDefinition)
        LOG.info("OnStrategyDefinitionProgress: requested symbol: %s GW symbol: %s status: %d",
            dispatchedDefinition.RequestString, dispatchedDefinition.Symbol, dispatchedDefinition.Status)
        if dispatchedDefinition.Status == constants.srsSuccess:
            self.eventStrategyIsReady.set()

    def OnOrderChanged(self, changeType, cqgOrder, oldProperties, cqgFill, cqgError):
        if cqgError is not None:
            error = win32com.client.Dispatch(cqgError)
            LOG.error("OnOrderChanged error: Code: %d Description: %s", error.Code, error.Description)
            return

        dispatchedOrder = win32com.client.Dispatch(cqgOrder)
        properties = win32com.client.Dispatch(dispatchedOrder.Properties)
        gwStatus = properties(constants.opGWStatus).Value
        quantity = properties(constants.opQuantity).Value
        instrument = properties(constants.opInstrumentName).Value
        limitPrice = properties(constants.opLimitPrice).Value

        LOG.info(
            "OnOrderChanged: change type: %d; GW status: %d; Quantity: %d; Instrument: %s; Price: %f;",
            changeType,
            gwStatus,
            quantity,
            instrument,
            limitPrice
        )

        if changeType != constants.ctChanged:
            return

        if gwStatus == constants.osInOrderBook:
            LOG.info("Strategy order is placed!")
            self.eventOrderPlaced.set()

    def OnInstrumentResolved(self, symbol, cqgInstrument, cqgError):
        if cqgError is not None:
            error = win32com.client.Dispatch(cqgError)
            LOG.error("OnInstrumentResolved error: Error code: %d Description: %s",
                      error.Code, error.Description)
            return

        self.instrument = cqgInstrument
        LOG.info("Instrument %s is resolved!", symbol)
        self.eventInstrumentIsReady.set()


if __name__ == "__main__":
    CELEnvironment.Start(StrategyPlacingSample)
